import torch
import torch.nn as nn
from networks import ObsEncoder, Sender, Receiver, ActionSelect
from utils import weights_init


class Agent(nn.Module):
    """
    Defines an agent module.
    """

    def __init__(self, obs_dim, msg_dim, num_actions, blind=False):
        """
        :param obs_dim: Dimension of output of ObsEncoder, also the dimension of hidden state of accumulator
        :param msg_dim: Dimension of output of Receiver
        :param num_actions: Number of actions that can be taken by this agent
        :param blind: Whether the agent is blind or not
        """
        super(Agent, self).__init__()

        self.num_actions = num_actions
        self.obs_dim = obs_dim
        self.msg_dim = msg_dim

        # Initialize if the agent is blind
        self.blind = blind

        # Initialize the networks
        self.obs_encoder = ObsEncoder(self.obs_dim)
        self.sender = Sender(self.obs_dim, self.msg_dim)
        self.receiver = Receiver(self.obs_dim, self.msg_dim)
        self.action_selector = ActionSelect(self.obs_dim, self.msg_dim, self.num_actions)

        # Initialize the hidden states
        self.acc_hidden = torch.zeros(1, self.obs_dim)
        self.recv_hidden = torch.zeros(1, self.msg_dim)

        # Initialize message queue
        self.messages = []
        self.new_messages = []

        # Initialize the neighbors
        self.neighbors = []

        # Initialize the weights
        self.apply(weights_init)

    def reset_agent(self):
        self.messages = []
        self.new_messages = []
        self.acc_hidden = torch.zeros(1, self.obs_dim)
        self.recv_hidden = torch.zeros(1, self.msg_dim)

    def add_message(self, msg):
        """
        :param msg: (1, msg_dim) Message sent by a neighbor
        """
        self.new_messages.append(msg)

    def reset_messages(self):
        self.messages = self.new_messages[:]
        self.new_messages = []

    def broadcast_message(self, msg):
        for neighbor in self.neighbors:
            if self.blind and neighbor.blind:   # Don't send message to blind agent if you are blind
                continue
            neighbor.add_message(msg)

    def add_neighbors(self, neighbor):
        """
        :param neighbor: An instance of class Agent
        """
        self.neighbors.append(neighbor)

    def accumulate_obs(self, obs):
        """
        Accumulates a single observation.
        :param obs: (3, 140, 140) observation at the current time step
        """
        self.acc_hidden = self.obs_encoder(obs.unsqueeze(0), self.acc_hidden)

    def forward(self, obs, evaluate=False):
        """
        :param obs: (3, 140, 140) observation at the current time-step
        :param evaluate: If true, then the model will be run in test mode
        :return: action, log_probs, out_msg
            action: (1,) action taken at the current time step
            log_prob: (1,) log probability of action taken at the current time step
            out_msg: (1, msg_dim) message sent by this node
        """
        # Process the observation
        self.acc_hidden = self.obs_encoder(obs.unsqueeze(0), self.acc_hidden)

        # Process the incoming messages
        if len(self.messages) > 0:
            messages = torch.cat(self.messages, dim=0)
            if self.blind:
                self.recv_hidden = self.receiver(torch.zeros(self.acc_hidden.size()), messages.unsqueeze(0),
                                                 self.recv_hidden)
            else:
                self.recv_hidden = self.receiver(self.acc_hidden, messages.unsqueeze(0), self.recv_hidden)

        # Process outgoing message
        out_msg = self.sender(self.acc_hidden, self.recv_hidden)
        self.broadcast_message(out_msg)

        # Take the action
        if self.blind:
            action, log_prob = self.action_selector(torch.zeros(self.acc_hidden.size()), self.recv_hidden, evaluate)
        else:
            action, log_prob = self.action_selector(self.acc_hidden, self.recv_hidden, evaluate)

        return action, log_prob, out_msg
