import torch.nn as nn
import torch
from torch.distributions import Gumbel, Categorical


class ObsEncoder(nn.Module):
    """
    Implements the observation encoder module where input observations are first passed through a convolutional neural
    network and then aggregated over time using a GRU.
    """

    def __init__(self, hidden_dim=64):
        """
        :param hidden_dim: Number of hidden units used by the GRU
        """
        super(ObsEncoder, self).__init__()
        self.feature_extractor = nn.Sequential(
            nn.Conv2d(in_channels=3, out_channels=16, kernel_size=10, stride=4, padding=1),
            nn.BatchNorm2d(num_features=16),
            nn.ReLU(),
            nn.Conv2d(in_channels=16, out_channels=32, kernel_size=9, stride=3, padding=1),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=9, stride=3, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2)
        )
        self.rnn = nn.GRUCell(input_size=64, hidden_size=hidden_dim)

    def forward(self, inputs, hidden_state):
        """
        :param inputs: (batch_size, 3, 140, 140) input observation
        :param hidden_state: (batch_size, hidden_dim) input hidden state of GRU
        :return: output (batch_size, hidden_dim) output hidden state of GRU
        """
        assert inputs.size(2) == 140, "Current setup uses 140 x 140 observations"
        assert inputs.size(2) == inputs.size(3), "Input has to be a square image"

        feats = self.feature_extractor(inputs)
        feats = feats.view(feats.size(0), -1)
        return self.rnn(feats, hidden_state)


class Sender(nn.Module):
    """
    Implements the sender module that takes the summary of input messages till the previous time-step along with the
    output of ObsEncoder for the current time step as input and produces the binary message to be sent as output
    """

    def __init__(self, obs_dim, msg_dim):
        """
        :param obs_dim: Dimension of the output of ObsEncoder
        :param msg_dim: Dimension of messages (number of bits)
        """
        super(Sender, self).__init__()
        self.msg_dim = msg_dim
        self.linear1 = nn.Linear(in_features=obs_dim + self.msg_dim, out_features=64)
        self.act1 = nn.ReLU()
        self.linear2 = nn.Linear(in_features=64, out_features=2 * self.msg_dim)
        self.gumbel_softmax = GumbelSoftmax()

    def forward(self, obs, messages, inv_temp=1.0):
        """
        :param obs: (batch_size, obs_dim) Input observation summary till the current time step. Output of ObsEncoder.
        :param messages: (batch_size, msg_dim) Input message summary till the previous time step. Output of Receiver.
        :param inv_temp: Scalar indicating inverse temperature value used for Gumbel-Softmax
        :return: out_messages (batch_size, msg_dim) Output messages (as binary vectors) for broadcasting
        """
        concat_embeds = torch.cat([obs, messages], dim=1)
        concat_embeds = self.linear2(self.act1(self.linear1(concat_embeds)))
        concat_embeds = concat_embeds.view(-1, self.msg_dim, 2)
        return self.gumbel_softmax(concat_embeds, inv_temp=inv_temp)


class Receiver(nn.Module):
    """
    Implements the receiver module that takes the input messages from all the neighboring agents along with the summary
    of observation at the current time step to generate a summary of input messages.
    """

    def __init__(self, obs_dim, msg_dim):
        """
        :param obs_dim: Dimension of the output of ObsEncoder
        :param msg_dim: Dimension of messages (number of bits)
        """
        super(Receiver, self).__init__()
        self.msg_dim = msg_dim
        self.query_linear = nn.Linear(in_features=obs_dim, out_features=self.msg_dim)
        self.key_linear = nn.Linear(in_features=self.msg_dim, out_features=self.msg_dim)
        self.softmax = nn.Softmax(dim=-1)
        self.rnn = nn.GRUCell(self.msg_dim, self.msg_dim)

    def forward(self, obs, input_messages, hidden_states):
        """
        :param obs: (batch_size, obs_dim) Output of the ObsEncoder network
        :param input_messages:  (batch_size, num_neighbors, msg_dim) Input messages from all neighbors
        :param hidden_states: (batch_size, msg_dim) Input hidden states for GRU
        :return: (batch_size, msg_dim) Pooled and aggregated message (aggregation done over time using GRU). Also the
                 output hidden state for GRU
        """
        num_neighbors = input_messages.size(1)

        # Process the key and query
        query = self.query_linear(obs).unsqueeze(1).repeat(1, num_neighbors, 1)
        key = self.key_linear(input_messages)   # Note that Linear layer automatically broadcasts the operation

        # Compute the attention weights
        attention_weights = self.softmax((query * key).sum(dim=2))

        # Compute the pooled messages
        attention_weights = attention_weights.unsqueeze(2).repeat(1, 1, self.msg_dim)
        pooled_msgs = (input_messages * attention_weights).sum(dim=1)

        # Aggregate the messages over time
        return self.rnn(pooled_msgs, hidden_states)


class ActionSelect(nn.Module):
    """
    Implements the action selection module which takes the current observation and pooled messages as input and
    chooses an action.
    """
    def __init__(self, obs_dim, msg_dim, num_actions):
        """
        :param obs_dim: Observation dimension (dimension of output of ObsEncoder)
        :param msg_dim: Message dimension (dimension of output of Receiver)
        :param num_actions: Number of actions
        """
        super(ActionSelect, self).__init__()
        self.linear1 = nn.Linear(obs_dim + msg_dim, 64)
        self.act1 = nn.ReLU()
        self.linear2 = nn.Linear(64, num_actions)
        self.log_softmax = nn.LogSoftmax(dim=-1)

    def forward(self, obs, messages, evaluate=False):
        """
        :param obs: (batch_size, obs_dim) Output of ObsEncoder from current time step
        :param messages: (batch_size, msg_dim) Output of Receiver
        :param evaluate: If true, then arg-max over action will be taken rather than sampling an action
        :return: actions, log_probs
            actions: (batch_size,) index of chosen actions
            log_probs: (batch_size,) Log probability of chosen action under the current policy
        """
        concat_embeds = torch.cat([obs, messages], dim=1)
        log_probs = self.log_softmax(self.linear2(self.act1(self.linear1(concat_embeds))))
        if evaluate:
            actions = torch.argmax(log_probs, dim=1)
        else:
            probs = torch.exp(log_probs)
            actions = Categorical(probs=probs).sample()
        log_probs = log_probs.gather(1, actions.view(-1, 1)).view(-1)
        return actions, log_probs


class GumbelSoftmax(nn.Module):
    """
    Implements the Gumbel-Softmax module
    """

    def __init__(self):
        super(GumbelSoftmax, self).__init__()
        self.gumbel = Gumbel(torch.tensor([0.0]), torch.tensor([1.0]))
        self.log_softmax = nn.LogSoftmax(dim=-1)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, inv_temp=1.0):
        """
        :param x: (batch_size, msg_dim, 2) Input on which Gumbel Softmax has to be applied. Will be applied on 2nd dim.
        :param inv_temp: Inverse of the temperature
        :return:
            y: (batch_size, msg_dim) binary vector
        """
        y = self.log_softmax(x)
        y = y + self.gumbel.sample(y.size()).squeeze(len(y.size()))
        return GumbelSoftmax.binarize(self.softmax(y * inv_temp))

    @staticmethod
    def binarize(x):
        """
        :param x: (batch_size, msg_dim, 2) output of Gumbel softmax
        :return:
            binary: (batch_size, msg_dim) Binary vector corresponding to x
        """
        idx = torch.max(x, dim=2)[1]
        x_hat = torch.zeros(x.size(0), x.size(1))
        for i in range(x.size(0)):
            x_hat[i, :] = x[i, :, :].gather(1, idx[i, :].view(-1, 1)).view(-1)
        binary = (idx.float() - x_hat).detach() + x_hat
        return binary
