"""
    Implements the Q-learning baseline using DQN.
"""

import os
import numpy as np
import torch
import torch.nn as nn
import utils
from torch.optim.rmsprop import RMSprop
import gc
from torch.nn.utils import clip_grad_value_


def soft_update(target, source):
    """
    :param target: Target neural network
    :param source: Source neural network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)


def hard_update(target, source):
    """
    :param target: Target neural network
    :param source: Source neural network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(param.data)


def weight_init(module):
    if isinstance(module, nn.Linear):
        nn.init.xavier_normal_(module.weight, gain=1)
        nn.init.constant_(module.bias, 0)


def reset_sumo():
    try:
        utils.end_sumo()    # Sometimes results in exception is SUMO is already closed
    except:
        pass

    try:
        os.system('killall -9 sumo-gui')    # Needed to avoid memory leak
    except:
        pass

    if os.path.exists('./maps/manhattan_1.rou.xml'):
        os.remove('./maps/manhattan_1.rou.xml')
    if os.path.exists('./maps/manhattan_1.rou.alt.xml'):
        os.remove('./maps/manhattan_1.rou.alt.xml')
    if os.path.exists('./maps/tripinfo.xml'):
        os.remove('./maps/tripinfo.xml')
    if os.path.exists('./maps/trips.trips.xml'):
        os.remove('./maps/trips.trips.xml')


def retrieve_reward(time, listlanes):
    if time == 0:
        return 0, 0, 0
    r1 = utils.get_queue_length(listlanes)
    r2 = utils.get_waiting_time(listlanes)
    r3 = utils.get_delay(listlanes)
    return r1, r2, r3


def read_saved_timesteps():
    filenames = []
    for _, _, filenames in os.walk(model_path):
        break
    steps = []
    for file in filenames:
        step_name, _ = file.split('_')
        step_name = int(step_name)
        if step_name not in steps:
            steps.append(step_name)
    return sorted(steps)


class DeepQNetwork(nn.Module):

    def __init__(self, num_actions):
        """
        :param num_actions: Number of actions that this agent can take
        """
        super(DeepQNetwork, self).__init__()

        self.num_actions = num_actions

        self.feature_extractor = nn.Sequential(
            nn.Conv2d(in_channels=3 * period, out_channels=32, kernel_size=10, stride=4, padding=1),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=9, stride=3, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(),
            nn.Conv2d(in_channels=64, out_channels=128, kernel_size=9, stride=3, padding=1),
            nn.BatchNorm2d(num_features=128),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2)
        )
        self.linear1 = nn.Linear(128, 64)
        self.act1 = nn.ReLU()
        self.linear2 = nn.Linear(64, self.num_actions)

        self.apply(weight_init)

    def forward(self, inputs):
        """
        :param inputs: (bs, 3 * period, 140, 140) input observations
        :return: q_vals (bs, num_actions) q-values corresponding to the observations
        """
        bs = inputs.size(0)
        return self.linear2(self.act1(self.linear1(self.feature_extractor(inputs).view(bs, -1))))

    def choose_action(self, q_vals, eps=0.05, test=False):
        """
        :param q_vals: (bs, num_action) Q-values
        :param eps: Probability with which random actions are taken while training
        :param test: Whether to run in test mode
        :return: actions (bs,) Chosen action index
        """
        idx = torch.argmax(q_vals, dim=1)
        if test:
            return idx

        rand_idx = torch.randint(0, self.num_actions, (q_vals.size(0),)).float()
        choose_rand = (torch.rand((q_vals.size(0),)) <= eps).float()
        idx = idx.float() * (torch.ones(choose_rand.size()) - choose_rand) + choose_rand * rand_idx

        return idx.long()


class ReplayBuffer:
    """
    Implements a replay buffer saving (s, a, r, s', done) tuple
    """

    def __init__(self, capacity=10000):
        """
        :param capacity: Maximum number of (s, a, r, s', done) that can be stored. The buffer is populated in a round
                         robin fashion.
        """
        self.capacity = capacity
        self.current_idx = 0
        self.num_examples = 0
        self.buffer = [None] * self.capacity

    def add(self, s, a, r, s_next, done):
        """
        :param s: (3 * period, 140, 140) Current state
        :param a: Scalar current action index
        :param r: Scalar reward
        :param s_next: (3 * period, 140, 140) Next state
        :param done: Whether it is the end of episode
        """
        self.buffer[self.current_idx] = (s, a, r, s_next, done)
        self.current_idx = (self.current_idx + 1) % self.capacity

        if self.num_examples < self.capacity:
            self.num_examples += 1

    def sample(self, bs=128):
        """
        :param bs: Number of examples to be sampled
        Returns: batch = [state, action, reward, next_state, done]
            state: (batch_size, 3 * period, 140, 140) current state
            action: (batch_size,) current action
            reward: (batch_size,) current reward
            next_state: (batch_size, 3 * period, 140, 140) next state
            done: (batch_size,) binary float vector
        """

        # Sample examples
        idx = np.random.randint(0, self.num_examples, size=(bs,)).tolist()
        batch = [self.buffer[i] for i in idx]

        # Get input in the correct format
        batch = list(map(np.stack, zip(*batch)))
        batch[-1] = batch[-1].astype(float)

        return batch


def run_episode():
    """
    Collects data worth a single episode in the replay buffer
    :return episode_rewards: A dictionary indexed by agents containing episode reward for each agent
    """

    # Generate random trips for the episode
    os.system('python randomTrips.py -n ' + net_file + ' -r ' + route_file)

    # Initialize SUMO
    utils.start_sumo()

    # Initialize the variables
    prev_state = dict([(agent, np.zeros((period * 3, 140, 140))) for agent in agents])
    curr_state = dict([(agent, np.zeros((period * 3, 140, 140))) for agent in agents])
    actions = dict([(agent, 0) for agent in agents])
    rewards = dict([(agent, 0) for agent in agents])

    # Start the episode
    counter = 0
    episode_rewards = dict([(agent, 0) for agent in agents])
    while counter < episode_length:

        # Get the current state
        for i in range(period):

            # Get the screenshot
            clock_time = utils.get_current_time()
            img = utils.get_snapshot(clock_time)

            # Accumulate the observations and rewards
            for agent in agents:
                curr_state[agent][i * 3: i * 3 + 3, :, :] = np.transpose(utils.crop_junctions(img, agent), (2, 0, 1))
                queue, wait, delay = retrieve_reward(counter, adjoining_lanes[agent])
                rewards[agent] += (-queue - wait - delay)

            counter += 1

        # Append items to replay buffer
        if counter > period:
            for agent in agents:
                replay_buffers[agent].add(prev_state[agent].copy().astype(np.float32), actions[agent], rewards[agent],
                                          curr_state[agent].copy().astype(np.float32), counter == episode_length)
                episode_rewards[agent] += rewards[agent]

        # Take actions for all agents and collect rewards
        for agent in agents:
            actions[agent] = agents[agent].choose_action(agents[agent](
                torch.from_numpy(curr_state[agent]).float().unsqueeze(0)), eps=epsilon).detach().item()
            utils.changeTrafficLight(agent, actions[agent])

        # Reset state
        for agent in agents:
            prev_state[agent] = curr_state[agent].copy()
            rewards[agent] = 0.0

    # End the simulator
    utils.end_sumo()

    return episode_rewards


def train():

    # Start training
    for agent in agents:

        # Don't train if replay buffer is not full enough
        if replay_buffers[agent].num_examples < batch_size:
            continue

        # Fetch the batch
        s, a, r, s_next, terminal = replay_buffers[agent].sample(batch_size)
        s = torch.from_numpy(s).float()
        a = torch.from_numpy(a).long()
        r = torch.from_numpy(r).float()
        s_next = torch.from_numpy(s_next).float()
        terminal = torch.from_numpy(terminal).float()

        # Compute Q(s, a)
        q_vals = agents[agent](s)
        q_vals = q_vals.gather(1, a.view(-1, 1)).view(-1)

        # Compute max_a' Q(s', a') and the target
        q_vals_next = agents_target[agent](s_next)
        action_idx = agents[agent].choose_action(q_vals_next, test=True)
        q_vals_next = q_vals_next.gather(1, action_idx.view(-1, 1)).view(-1)
        target = (r + (torch.ones(terminal.size()) - terminal) * gamma * q_vals_next).detach()

        # Optimize
        optimizers[agent].zero_grad()
        loss = ((q_vals - target) ** 2).mean()
        loss.backward()
        clip_grad_value_(agents[agent].parameters(), 1.0)
        optimizers[agent].step()

        # Update the target Q-network
        soft_update(agents_target[agent], agents[agent])

    # Run an episode
    return run_episode()


def test_episode():
    """
    Runs Q-learning in the test mode
    :return: episode_rewards: A dictionary indexed by agents containing episode reward for each agent
    """
    # Generate random trips for the episode
    os.system('python randomTrips.py -n ' + net_file + ' -r ' + route_file)

    # Initialize SUMO
    utils.start_sumo()

    # Initialize the variables
    curr_state = dict([(agent, np.zeros((period * 3, 140, 140))) for agent in agents])
    episode_rewards = dict([(agent, 0) for agent in agents])

    # Start the episode
    counter = 0
    while counter < episode_length:

        # Get the current state
        for i in range(period):

            # Get the screenshot
            clock_time = utils.get_current_time()
            img = utils.get_snapshot(clock_time)

            # Accumulate the observations and rewards
            for agent in agents:
                curr_state[agent][i * 3: i * 3 + 3, :, :] = np.transpose(utils.crop_junctions(img, agent), (2, 0, 1))
                queue, wait, delay = retrieve_reward(counter, adjoining_lanes[agent])
                episode_rewards[agent] += (-queue - wait - delay)

            counter += 1

        # Take actions for all agents and collect rewards
        for agent in agents:
            action = agents[agent].choose_action(agents[agent](
                torch.from_numpy(curr_state[agent]).float().unsqueeze(0)), test=True).detach().item()
            utils.changeTrafficLight(agent, action)

    # End the simulator
    utils.end_sumo()

    return episode_rewards


if __name__ == '__main__':
    # Define control variables
    net_file = './maps/manhattan_1.net.xml'
    route_file = './maps/manhattan_1.rou.xml'
    learning_rate = 1e-4
    period = 5
    episode_length = 200
    epsilon = 1.0
    batch_size = 128
    gamma = 0.9
    tau = 0.01
    num_episodes = 1000
    model_path = './model_q'
    train_model = True

    # Get the neighbors of all agents in the network
    neighbours = utils.find_neigbours(net_file)
    num_agents = len(neighbours)

    # Get the lanes connected to all traffic junctions
    adjoining_lanes = utils.find_incoming_lanes(net_file)

    if train_model:
        # Start logging
        log_file = open('log_q.txt', 'w')

        # Initialize the agents
        agents = dict([(agent, DeepQNetwork(utils.get_degree(agent))) for agent in neighbours])
        agents_target = dict([(agent, DeepQNetwork(utils.get_degree(agent))) for agent in neighbours])
        for agent in agents:
            hard_update(agents_target[agent], agents[agent])

        # Initialize the optimizers
        optimizers = dict([(agent, RMSprop(agents[agent].parameters(), lr=learning_rate)) for agent in agents])

        # Initialize replay buffers
        replay_buffers = dict([(agent, ReplayBuffer(800)) for agent in agents])

        # Start training
        for episode_idx in range(num_episodes):
            print('_________________________________________________________________________________________\n')
            try:
                ep_rewards = train()
                avg_reward = sum([ep_rewards[x] for x in ep_rewards]) / num_agents
                print('Episode:', episode_idx, 'Avg. Reward:', avg_reward)
                log_file.write('Episode: ' + str(episode_idx) + ' Avg. Reward: ' + str(avg_reward) + "\n")
                epsilon = max(0.99 * epsilon, 0.05)
            except Exception as exception:
                print('Error in episode:', episode_idx, 'Exception:', exception)
                log_file.write('Error in episode ' + str(episode_idx) + '\n')
                reset_sumo()

            log_file.flush()

            if episode_idx % 10 == 0:
                for agent in agents:
                    with open(os.path.join(model_path, str(episode_idx) + '_' + agent), 'wb') as f:
                        torch.save(agents[agent].state_dict(), f)

            gc.collect()
    else:
        # Start logging
        log_file = open('log_q_test.txt', 'w')

        # Read the saved model time-steps
        time_steps = read_saved_timesteps()

        # Initialize the agents
        agents = dict([(agent, DeepQNetwork(utils.get_degree(agent))) for agent in neighbours])

        # Start testing
        for step in time_steps:

            # Load the agent q-networks
            for agent in agents:
                agents[agent].load_state_dict(torch.load(os.path.join(model_path, str(step) + '_' + agent)))

            # Start testing
            for episode_idx in range(5):
                print('_________________________________________________________________________________________\n')
                try:
                    ep_rewards = test_episode()
                    avg_reward = sum([ep_rewards[x] for x in ep_rewards]) / num_agents
                    print('Step:', step, 'Episode:', episode_idx, 'Avg. Reward:', avg_reward)
                    log_file.write('Step: ' + str(step) + ' Episode: ' + str(episode_idx) + ' Avg. Reward: ' +
                                   str(avg_reward) + "\n")
                except:
                    print('Step:', step, 'Error in episode:', episode_idx)
                    log_file.write('Step: ' + str(step) + ' Error in episode ' + str(episode_idx) + '\n')
                    reset_sumo()

                log_file.flush()
