import os
import torch
from arguments import arguments
import utils
from agent import Agent
from tensorboardX import SummaryWriter


# Control variables
model_path = 'model'
args = arguments()
log_file = open('log_test.txt', 'w')


def reset_sumo():
    utils.end_sumo()
    if os.path.exists('./maps/manhattan_1.rou.xml'):
        os.remove('./maps/manhattan_1.rou.xml')
    if os.path.exists('./maps/manhattan_1.rou.alt.xml'):
        os.remove('./maps/manhattan_1.rou.alt.xml')
    if os.path.exists('./maps/tripinfo.xml'):
        os.remove('./maps/tripinfo.xml')
    if os.path.exists('./maps/trips.trips.xml'):
        os.remove('./maps/trips.trips.xml')


def retrieve_reward(time, listlanes):
    if time == 0:
        return 0, 0, 0
    r1 = utils.get_queue_length(listlanes)
    r2 = utils.get_waiting_time(listlanes)
    r3 = utils.get_delay(listlanes)
    # r4 = utils.get_emergency_decel(listlanes)
    return r1, r2, r3


def read_saved_timesteps():
    filenames = []
    for _, _, filenames in os.walk(model_path):
        break
    steps = []
    for file in filenames:
        step_name, _ = file.split('_')
        step_name = int(step_name)
        if step_name not in steps:
            steps.append(step_name)
    return sorted(steps)


def save_embeddings(msg, action, agent_id, episode, policy_number):
    num_time_steps = len(msg)
    writer.add_embedding(msg,
                         metadata=[agent_id] * num_time_steps,
                         tag='msg_' + agent_id + '_EP_' + str(episode) + '_' + str(policy_number))
    writer.add_embedding(action,
                         metadata=[agent_id] * num_time_steps,
                         tag='action_' + agent_id + '_EP_' + str(episode) + '_' + str(policy_number))


def test():
    # Generate trips
    os.system('python randomTrips.py -n ' + args.net_file + ' -r ' + args.route_file)

    # Initialize the rewards
    agent_rewards = dict([(agent, []) for agent in agents])

    # Start the simulator
    utils.start_sumo()

    # Reset the agents
    for agent in agents:
        agents[agent].reset_agent()

    # Start the episode
    messages = dict([(agent, []) for agent in agents])
    actions = dict([(agent, []) for agent in agents])
    for time in range(args.episode_length):
        # Get the screenshot
        clock_time = utils.get_current_time()
        img = utils.get_snapshot(clock_time)

        # Reset agent messages if needed
        if time % args.period == 0:
            for agent in agents:
                agents[agent].reset_messages()

        for agent in agents:
            agents[agent].eval()

            # Get the observations
            cropped_img = torch.tensor(utils.crop_junctions(img, agent))
            cropped_img = cropped_img.permute(2, 0, 1)

            # Use the observation
            if time % args.period != 0:
                agents[agent].accumulate_obs(cropped_img)
                queue, wait, delay = retrieve_reward(time, adjoining_lanes[agent])
                agent_rewards[agent][-1] += (-queue - wait - delay)
                continue

            # Take the action and accumulate log probabilities
            action, _, out_msg = agents[agent](cropped_img, evaluate=True)
            utils.changeTrafficLight(agent, action.item())

            # Collect the rewards
            queue, wait, delay = retrieve_reward(time, adjoining_lanes[agent])
            agent_rewards[agent].append(-queue - wait - delay)

            # Save the data
            messages[agent].append(out_msg.detach().cpu().tolist()[0])
            actions[agent].append([action])

    # Save the embeddings
    for agent in agents:
        save_embeddings(messages[agent], actions[agent], agent, episode_idx, step)

    # End the simulator
    utils.end_sumo()

    # Compute the losses
    for agent in agents:
        agent_rewards[agent] = torch.tensor(agent_rewards[agent]).sum().item()

    return agent_rewards


if __name__ == '__main__':

    # Initialize the summary writer
    writer = SummaryWriter()

    # Get the neighbors of all agents in the network
    neighbours = utils.find_neigbours(args.net_file)
    num_agents = len(neighbours)

    # Get the lanes connected to all traffic junctions
    adjoining_lanes = utils.find_incoming_lanes(args.net_file)

    # Read the saved model time-steps
    time_steps = read_saved_timesteps()

    # Initialize the agents
    agents = dict([(agent, Agent(obs_dim=args.hid_dim, msg_dim=args.msg_dim, num_actions=utils.get_degree(agent)))
                   for agent in neighbours])

    # Add neighbors
    for agent in agents:
        for nbr in neighbours[agent]:
            agents[agent].add_neighbors(agents[nbr])

    # Start testing
    for step in time_steps:

        # Load the agent profiles
        for agent in agents:
            agents[agent].load_state_dict(torch.load(os.path.join(model_path, str(step) + '_' + agent)))

        # Start testing
        for episode_idx in range(5):
            print('_________________________________________________________________________________________\n')
            try:
                episode_rewards = test()
                avg_reward = sum([episode_rewards[x] for x in episode_rewards]) / num_agents
                print('Step:', step, 'Episode:', episode_idx, 'Avg. Reward:', avg_reward)
                log_file.write('Step: ' + str(step) + ' Episode: ' + str(episode_idx) + ' Avg. Reward: ' +
                               str(avg_reward) + "\n")
            except:
                print('Step:', step, 'Error in episode:', episode_idx)
                log_file.write('Step: ' + str(step) + ' Error in episode ' + str(episode_idx) + '\n')
                reset_sumo()

            log_file.flush()

log_file.close()
