import os
import torch
from torch.optim.rmsprop import RMSprop
from arguments import arguments
import utils
from agent import Agent
from torch.nn.utils import clip_grad_value_
import gc


# Control variables
model_path = 'model'
args = arguments()
log_file = open('log.txt', 'w')


def reset_sumo():
    utils.end_sumo()
    if os.path.exists('./maps/manhattan_1.rou.xml'):
        os.remove('./maps/manhattan_1.rou.xml')
    if os.path.exists('./maps/manhattan_1.rou.alt.xml'):
        os.remove('./maps/manhattan_1.rou.alt.xml')
    if os.path.exists('./maps/tripinfo.xml'):
        os.remove('./maps/tripinfo.xml')
    if os.path.exists('./maps/trips.trips.xml'):
        os.remove('./maps/trips.trips.xml')


def retrieve_reward(time, listlanes):
    if time == 0:
        return 0, 0, 0
    r1 = utils.get_queue_length(listlanes)
    r2 = utils.get_waiting_time(listlanes)
    r3 = utils.get_delay(listlanes)
    # r4 = utils.get_emergency_decel(listlanes)
    return r1, r2, r3


def compute_loss(agent_rewards, log_probs):
    """
    :param agent_rewards: A dictionary indexed by agent id. Each entry corresponds to a torch vector of rewards received
                          by the agent at each time step when action was taken.
    :param log_probs: A dictionary indexed by agent id. Each entry is torch vector containing log probability of the
                      action chosen at each time step.
    :return: losses
        losses: A dictionary indexed by agent id. Each entry corresponds to the loss to be minimized by the agent.
    """
    losses = dict()

    # Compute loss for each agent
    for agent in agent_rewards:

        # Compute cumulative rewards
        rewards = agent_rewards[agent].clone()
        n = rewards.size(0)
        for i in range(n - 2, -1, -1):
            rewards[i] = rewards[i] + args.gamma * rewards[i + 1]

        # Compute loss
        losses[agent] = -((rewards - rewards_mean[agent]) * log_probs[agent]).mean()

        # Update the baseline
        rewards_mean[agent] = 0.9 * rewards_mean[agent] + 0.1 * rewards.detach()

    return losses


def train():
    """
    Returns: rewards
        losses: Loss incurred by all agents. A dictionary indexed by agent id.
    """

    # Generate random trips for the episode
    os.system('python randomTrips.py -n ' + args.net_file + ' -r ' + args.route_file)

    # Initialize variables
    agent_rewards = dict([(agent, []) for agent in agents])
    agent_log_probs = dict([(agent, []) for agent in agents])

    # Initialize SUMO
    utils.start_sumo()

    # Reset the agents
    for agent in agents:
        agents[agent].reset_agent()

    # Start the episode
    for time in range(args.episode_length):
        # Get the screenshot
        clock_time = utils.get_current_time()
        img = utils.get_snapshot(clock_time)

        # Reset agent messages if needed
        if time % args.period == 0:
            for agent in agents:
                agents[agent].reset_messages()

        for agent in agents:
            agents[agent].train()

            # Get the observations
            cropped_img = torch.tensor(utils.crop_junctions(img, agent))
            cropped_img = cropped_img.permute(2, 0, 1)

            # Use the observation
            if time % args.period != 0:
                agents[agent].accumulate_obs(cropped_img)
                queue, wait, delay = retrieve_reward(time, adjoining_lanes[agent])
                agent_rewards[agent][-1] += (-queue - wait - delay)
                continue

            # Take the action and accumulate log probabilities
            action, log_prob, _ = agents[agent](cropped_img)
            agent_log_probs[agent].append(log_prob)
            utils.changeTrafficLight(agent, action.item())

            # Collect the rewards
            queue, wait, delay = retrieve_reward(time, adjoining_lanes[agent])
            agent_rewards[agent].append(-queue - wait - delay)

    # End the simulator
    utils.end_sumo()

    # Compute the losses
    for agent in agents:
        agent_rewards[agent] = torch.tensor(agent_rewards[agent])
        agent_log_probs[agent] = torch.cat([x.view(-1, 1) for x in agent_log_probs[agent]], dim=1).squeeze()
    losses = compute_loss(agent_rewards, agent_log_probs)

    # Train the model
    for optim in optimizers.values():
        optim.zero_grad()

    agent_loss = torch.tensor(0.0)
    for agent in agents:
        agent_loss += losses[agent]
    agent_loss.backward()

    for agent in agents:
        clip_grad_value_(agents[agent].parameters(), 1.0)

    for optim in optimizers.values():
        optim.step()

    # Compute rewards to return
    for agent in agents:
        agent_rewards[agent] = agent_rewards.get(agent).sum().item()

    return agent_rewards


if __name__ == '__main__':

    # Create the required folders if necessary
    if not os.path.exists('./snapshots'):
        os.mkdir('snapshots')
    if not os.path.exists(model_path):
        os.mkdir(model_path)

    # Get the neighbors of all agents in the network
    neighbours = utils.find_neigbours(args.net_file)
    num_agents = len(neighbours)

    # Get the lanes connected to all traffic junctions
    adjoining_lanes = utils.find_incoming_lanes(args.net_file)

    # Initialize the agents
    if args.shared_policy:
        networks = dict([(utils.get_degree(agent), None) for agent in neighbours])
        networks = dict([(degree, Agent(obs_dim=args.hid_dim, msg_dim=args.msg_dim, num_actions=degree))
                         for degree in networks])   # Blind agents not supported when shared policy is used
        agents = dict([(agent, networks[utils.get_degree(agent)]) for agent in neighbours])
    else:
        blind_agents = args.blind_agents.split(',')
        agents = dict([(agent, Agent(obs_dim=args.hid_dim, msg_dim=args.msg_dim, num_actions=utils.get_degree(agent),
                                     blind=agent in blind_agents))
                       for agent in neighbours])

    # Initialize the optimizers
    optimizers = dict([(agent, RMSprop(agents[agent].parameters(), lr=args.learning_rate)) for agent in neighbours])

    # Initialize running mean of rewards
    rewards_mean = dict([(agent, torch.zeros(args.episode_length // args.period,)) for agent in agents])

    # Add neighbors
    for agent in agents:
        for nbr in neighbours[agent]:
            agents[agent].add_neighbors(agents[nbr])

    # Start training
    for episode_idx in range(args.num_episodes):
        print('_________________________________________________________________________________________\n')
        try:
            episode_rewards = train()
            avg_reward = sum([episode_rewards[x] for x in episode_rewards]) / num_agents
            print('Episode:', episode_idx, 'Avg. Reward:', avg_reward)
            log_file.write('Episode: ' + str(episode_idx) + ' Avg. Reward: ' + str(avg_reward) + "\n")
        except:
            print('Error in episode:', episode_idx)
            log_file.write('Error in episode ' + str(episode_idx) + '\n')
            reset_sumo()

        log_file.flush()

        if episode_idx % 10 == 0:
            for agent in agents:
                with open(os.path.join(model_path, str(episode_idx) + '_' + agent), 'wb') as f:
                    torch.save(agents[agent].state_dict(), f)

        gc.collect()

log_file.close()
