import os
import sys
import optparse
import numpy as np
import pylab as plt
import torch
from collections import defaultdict

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")

import sumolib
from sumolib import checkBinary
import traci
import torch.nn as nn


# for 4 lane junction
phase0 = "rrrr rrGG rrrr rrGG".replace(" ", "")
phase1 = "GGrr rrrr GGrr rrrr".replace(" ", "")
phase2 = "rrGG rrrr rrGG rrrr".replace(" ", "")
phase3 = "rrrr GGrr rrrr GGrr".replace(" ", "")

# for 3 lane junction
phase4 = "GGG rrr rrr".replace(" ", "")
phase5 = "rrr GGG rrr".replace(" ", "")
phase6 = "rrr rrr GGG".replace(" ", "")

controlSignal1 = (phase0, phase1, phase2, phase3)
controlSignal2 = (phase4, phase5, phase6)
Signal = [controlSignal1, controlSignal2]

degree_map = {u'0+0': 3, u'0+1': 3, u'0+2': 3,
              u'1+0': 4, u'1+1': 4, u'1+2': 4, u'1+3': 4,
              u'2+1': 3, u'2+2': 3, u'2+3': 3}

stats = {}


def start_sumo():
    # check binary
    sumo_binary = checkBinary('sumo-gui')

    # traci starts sumo as a subprocess and then this script connects and runs
    traci.start([sumo_binary, "-c", "sumo_config.sumocfg", "--tripinfo-output", "./maps/tripinfo.xml",
                 "--log", "sumo_log.txt", "--start", "--quit-on-end", "--window-size", "1855,1004"])


def end_sumo():
    traci.close(False)


def get_current_time():
    return traci.simulation.getTime() / 1000


def find_neigbours(net_file):
    # set junctions without traffic lights to none
    neighbours = defaultdict(list)
    net = sumolib.net.readNet(net_file)
    edges = net._edges
    for i in range(len(edges)):
        if edges[i].getToNode().getID()[-4:] != 'none' and \
                edges[i].getFromNode().getID()[-4:] != 'none':
            from_id = edges[i].getFromNode().getID()
            neighbours[from_id] += [edges[i].getToNode().getID()]
    return neighbours


def find_incoming_lanes(net_file):
    # set junctions without traffic lights to none
    lanes = defaultdict(list)
    net = sumolib.net.readNet(net_file)
    edges = net._edges
    for i in range(len(edges)):
        if edges[i].getToNode().getID()[-4:] != 'none':
            junction = edges[i].getToNode().getID()
            lanes[junction] += [edges[i].getID() + '_0']
            lanes[junction] += [edges[i].getID() + '_1']
    return lanes


def get_degree(ID):
    return degree_map[ID]


def action_space(ID):
    degree = get_degree(ID)
    if degree == 4:
        return Signal[0]
    else:
        return Signal[1]


def get_ID_list():
    light_id = traci.trafficlight.getIDList()
    return light_id
  
def getTrafficLight_phase(agent):
    return traci.trafficlight.getRedYellowGreenState(agent)

def changeTrafficLight(agent, phase):
    degree = degree_map[agent]
    index = 0 if degree == 4 else 1
    nxt_phase = Signal[index][phase]
    traci.trafficlight.setRedYellowGreenState(agent, nxt_phase)


def crop_junctions(img, ID):
    locs = {
        '0+0': (0, 140, 307, 447),
        '0+1': (0, 140, 670, 810),
        '0+2': (0, 140, 1034, 1174),
        '1+0': (344, 484, 307, 447),
        '1+1': (344, 484, 670, 810),
        '1+2': (344, 484, 1034, 1174),
        '1+3': (344, 484, 1403, 1543),
        '2+1': (680, 820, 670, 810),
        '2+2': (680, 820, 1034, 1174),
        '2+3': (680, 820, 1403, 1543)
    }
    x1, x2, y1, y2 = locs[ID]
    return img[x1:x2, y1:y2, :3]


def get_snapshot(i):
    filename = "snapshots/" + str(i) + ".png"
    traci.gui.screenshot("View #0", filename)
    traci.simulationStep()
    img = plt.imread(filename)
    return img


def get_weight_matrix(mapper, neighbours):
    mat = np.zeros((len(neighbours), len(neighbours)))
    for agent, adjacent in neighbours.items():
        mat[mapper[agent], list(map(mapper.__getitem__, adjacent))] = \
            1. / len(adjacent)

    return mat


def get_vehicle_id_entering(lane):
    return traci.lane.getLastStepVehicleIDs(lane)


def get_queue_length(listLanes):
    queue_length = 0
    for lane in listLanes:
        queue_length += traci.lane.getLastStepHaltingNumber(lane)
    return queue_length


def get_waiting_time(listLanes):
    wait_time = 0
    for lane in listLanes:
        wait_time += traci.lane.getWaitingTime(lane)
    return wait_time


def get_delay(listLanes):
    delay = 0
    for lane in listLanes:
        delay += 1 - traci.lane.getLastStepMeanSpeed(lane) / traci.lane.getMaxSpeed(lane)
    return delay


def get_emergency_decel(listLanes):
    decel = 0
    for lane in listLanes:
        try:
            vehicle_id_entering_list = get_vehicle_id_entering(lane)
            decel += traci.vehicle.getEmergencyDecel(vehicle_id_entering_list)
        except:
            continue

    return decel


def print_stats(reward1, episode):
    # print(reward1.values())
    reward1 = dict((k, np.array(v).mean()) for k, v in list(reward1.items()))
    print(reward1)
    print(episode, np.array(list(reward1.values())).sum())


def save_model(models, PATH, episode):
    for ID in models.keys():
        torch.save(models[ID].state_dict(), PATH + str(episode)+ '_' + ID)
    return


def load_model(model, PATH):
    model.load_state_dict(torch.load(PATH))
    return model


# Bucket the messages for fixed communication protocol
def bucket(r1, r2, r3):
    msg = []

    if -r1 <= 4:
        msg += [0]
    else:
        msg += [1]

    if -r2 <= 100:
        msg += [0]
    else:
        msg += [1]

    if -r3 <= 2:
        msg += [0]
    else:
        msg += [1]

    return msg


def weights_init(module):
    if isinstance(module, nn.Linear):
        nn.init.xavier_uniform_(module.weight, gain=1)
        nn.init.constant_(module.bias, 0)
